#include <tiffio.h>
#include <stdio.h>
#include <stdlib.h>
#include "fitsio.h"
#include <string.h>

#define BUF_SIZE 10240

void print_help(){
 fprintf(stdout,"tiff2fits v2.1 (c) Alexander Lebedev & Kirill Sokolovsky\nUsage: ./tiff2fits [OPTIONS] input_file.tif output_file.fit\n");
 fprintf(stdout," OPTIONS:\n -c NUMBER or --channel NUMBER  Take colour chanel:\n                                0 - RED, 1 - GREEN (default), 2 - BLUE\n -h        or --help            Print this message\n -i        or --invert          Invert image (make white stars on black background)\n");
 return;
}

int main(int argc, char **argv)
{
  fitsfile *fptr;                                   
  int status = 0;                                    
  char buf[BUF_SIZE];
  TIFF *tif;
  uint32 width, height, i, j, p, memsize = 50000000, hmax, shift;               //memsize - size of memory bufer. 50000000
  tsize_t lsize;
  uint16 sample_per_pix, sample_size, planar_config, channel = 1; //channel - channel of image
  tdata_t *line_buf;
  long naxes[2];
  unsigned short *image;
  int ii;
  /* special parameters */
  int  negative=0;
  
  
  if(argc < 3)
    { 
//      fprintf(stderr, "Poshel na xuy\n");
      print_help();
      exit(0);
    }
  for(ii=1;ii<argc;ii++){
   if( 0==strcasecmp(argv[ii],"-h") || 0==strcasecmp(argv[ii],"--help") )
    print_help();
   if( 0==strcasecmp(argv[ii],"-c") || 0==strcasecmp(argv[ii],"--channel") && ii+1<argc )
    channel=atoi(argv[ii+1]);  
   if( 0==strcasecmp(argv[ii],"-b") && ii+1<argc )
    memsize=atoi(argv[ii+1]);
   if( 0==strcasecmp(argv[ii],"-i") || 0==strcasecmp(argv[ii],"--invert") )
    negative=1;
  }  
  if( negative==1 )
   fprintf(stderr, "Negative image will be produced\n");
  if( channel==0 )
   fprintf(stderr, "Taking RED channel\n");
  if( channel==1 )
   fprintf(stderr, "Taking GREEN channel\n");
  if( channel==2 )
   fprintf(stderr, "Taking BLUE channel\n");
  
  tif = TIFFOpen(argv[argc-2],"r");
  if(tif == NULL) 
    {
      fprintf(stderr, "Can't open file %s\n", argv[argc-2]);
      exit(0);
    }
  
  if(TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &width)==0)
    {
      fprintf(stderr, "Can't get TIFFTAG_IMAGEWIDTH\n");
      exit(0);
    }
  if(TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &height)==0)
    {
      fprintf(stderr, "Can't get TIFFTAG_IMAGELENGTH\n");
      exit(0);
    }
  if(TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, &sample_per_pix)==0)
    {
      fprintf(stderr, "Can't get TIFFTAG_SAMPLESPERPIXEL\n");
      exit(0);
    }                   
  if(TIFFGetField(tif, TIFFTAG_PLANARCONFIG, &planar_config)==0)
    {
      fprintf(stderr, "Can't get TIFFTAG_PLANARCONFIG\n");
      exit(0);
    }
  
  if(planar_config != 1)
    {
      fprintf(stderr, "Planar Config = %d. It can work incorrectly, continue? n/yes: ", planar_config);
      fgets(buf, BUF_SIZE, stdin);
      if(strncmp(buf, "yes", BUF_SIZE) != 0)exit(0); 
    }                    
  
  lsize = TIFFScanlineSize(tif);
  if(lsize % width != 0)
    {
      fprintf(stderr, "Size of one line in bytes is not divide by width");
      exit(0);
    }
  
  sample_size = lsize/width/sample_per_pix;
  if(sample_size != 1 && sample_size != 2)
    {
      fprintf(stderr, "Not support sample\n");
      exit(0);
    }  
  
  if(channel >= sample_per_pix )
    {
      fprintf(stderr, "Channel must be < %d and >= 0\n", sample_per_pix );
      exit(0);
    }

  fprintf(stderr, "----TIFF file info:----\n width = %d\n height = %d\n sample_per_pix = %d\n line_size = %ld\n sample_size = %d\n planar_config = %d\n-----------------------\n", width, height, sample_per_pix, lsize, sample_size, planar_config);
  
  fits_create_file(&fptr, argv[argc-1], &status);
  if(status == FILE_NOT_CREATED)
    {
      fprintf(stderr, "Can't creat file %s\n", argv[argc-1]);
      exit(0);
    }
  
  naxes[0] = width;
  naxes[1] = height;
  
  fits_create_img(fptr, USHORT_IMG, 2, naxes, &status);
  line_buf = _TIFFmalloc(lsize);
  if(line_buf == NULL)
    {
      fprintf(stderr, "Can't run _TIFFmalloc(%ld)\n", lsize);
      exit(0);
    }
  
  
  hmax = memsize/width/sizeof(unsigned short);
  if(hmax > height)hmax = height;
  if(hmax == 0)
    {
      fprintf(stderr, "Very litle memsize\n");
      exit(0);
    }
  fprintf(stderr, "Locating %ld bytes\n", hmax*width*sizeof(unsigned short));
  image = malloc(hmax*width*sizeof(unsigned short));
  if( image == NULL )
    {
      fprintf(stderr, "No memory\n");
      exit(0);
    }
  
  for(i = 0;i < height;)
    {
      for(j = 0; (j < hmax) && (i < height); j++, i++)
	{
	  if(TIFFReadScanline(tif, line_buf, i, 0) == -1)
	    fprintf(stderr, "Can't read line of tiff, line = %d\n", i);
	  if(sample_size == 1)
	    {       
	      for(p = 0; p < width; p++)
		{
		  shift = p*sample_per_pix+channel;
		  image[j*width+p] = ((unsigned char *)line_buf)[shift];
		  if( negative==1 )
		   image[j*width+p]=255-image[j*width+p];
		}
	    }
	  else if(sample_size == 2)
	    {
	      for(p = 0; p < width; p++)
		{              
		  shift = (p*sample_per_pix+channel)*sample_size;
		  image[j*width+p] = *((unsigned short *)((unsigned char *)line_buf+shift));
		  if( negative==1 )
		   image[j*width+p]=65535-image[j*width+p];
		}
	    }
	}
      fits_write_img(fptr, TUSHORT, (i-j)*width+1, width*j, image, &status);
    }  
  
  
  free(image);
  _TIFFfree(line_buf);
  fits_close_file(fptr, &status);
  fits_report_error(stderr, status);
  TIFFClose(tif);
  return 0;
}

